TERRAFORM_VERSION = "1.0.5"

# 04-kubernetes-ingress
DEFAULT_TRAEFIK_IMAGE_TAG = "2.9.1"

HIGHEST_SUPPORTED_K8S_VERSION = "1.23.5"

DEFAULT_NEBARI_DASK_VERSION = "2023.1.1"
DEFAULT_NEBARI_IMAGE_TAG = "2023.1.1"

DEFAULT_CONDA_STORE_IMAGE_TAG = "v0.4.12"

LATEST_SUPPORTED_PYTHON_VERSION = "3.10"
